import ctypes as ct
import argparse
import serial
import os
import subprocess

CRSF_CRC_POLY = 0xD5
CRCLEN = 256
FRAME_SIZE = 7
uint8_t = ct.c_uint8

VOXL_PLATFORM = False
try:
	ret = subprocess.check_output(['voxl-platform'])
	VOXL_PLATFORM = True
except FileNotFoundError:
	VOXL_PLATFORM = False

# Detect if we are on a voxl2 or host computer
if VOXL_PLATFORM:
    import voxl_serial
    BASE_DIR = '/usr/share/modalai/voxl-elrs'
    host = "voxl2"
else:
    BASE_DIR = '/usr/share/modalai/voxl-elrs'
    host = "ubuntu"

class GENERIC_CRC8:
    def __init__(self, poly= 0xD5):
        self.crcpoly = poly
        self.crc8tab = [0, 213, 127, 170, 254, 43, 129, 84, 41, 252, 86, 131, 215, 2, 168, 125, 82, 135,
                        45, 248, 172, 121, 211, 6, 123, 174, 4, 209, 133, 80, 250, 47, 164, 113, 219, 14, 
                        90, 143, 37, 240, 141, 88, 242, 39, 115, 166, 12, 217, 246, 35, 137, 92, 8, 221, 
                        119, 162, 223, 10, 160, 117, 33, 244, 94, 139, 157, 72, 226, 55, 99, 182, 28, 201, 
                        180, 97, 203, 30, 74, 159, 53, 224, 207, 26, 176, 101, 49, 228, 78, 155, 230, 51, 
                        153, 76, 24, 205, 103, 178, 57, 236, 70, 147, 199, 18, 184, 109, 16, 197, 111, 186, 
                        238, 59, 145, 68, 107, 190, 20, 193, 149, 64, 234, 63, 66, 151, 61, 232, 188, 105, 
                        195, 22, 239, 58, 144, 69, 17, 196, 110, 187, 198, 19, 185, 108, 56, 237, 71, 146, 
                        189, 104, 194, 23, 67, 150, 60, 233, 148, 65, 235, 62, 106, 191, 21, 192, 75, 158, 
                        52, 225, 181, 96, 202, 31, 98, 183, 29, 200, 156, 73, 227, 54, 25, 204, 102, 179, 231,
                        50, 152, 77, 48, 229, 79, 154, 206, 27, 177, 100, 114, 167, 13, 216, 140, 89, 243, 
                        38, 91, 142, 36, 241, 165, 112, 218, 15, 32, 245, 95, 138, 222, 11, 161, 116, 9, 220,
                        118, 163, 247, 34, 136, 93, 214, 3, 169, 124, 40, 253, 87, 130, 255, 42, 128, 85, 1, 
                        212, 126, 171, 132, 81, 251, 46, 122, 175, 5, 208, 173, 120, 210, 7, 83, 134, 44, 249 ]

    def calc(self, data, length , crc : uint8_t =0):
        for i in range(length):
            crc= self.crc8tab[crc ^ data[i]]
        return crc

def str2us(microseconds: int) -> (uint8_t, uint8_t):
    if microseconds < 256:
        return microseconds, 0
    else:
        return (microseconds >> 8),(microseconds & 0xFF)

def parse_args() -> dict:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-c', '--channel', help='Channel to set pwm', required=True, type=str)
    parser.add_argument('-s', '--microseconds', help='Microseconds ON', required=False, type=str)
    parser.add_argument('-d', '--duty', help='Duty cycle %', required=False, type=str)
    args = vars(parser.parse_args())
    
    if args['microseconds'] == None and args['duty'] == None:
        print("[ERROR] Must specify microseconds or duty cycle.")
        exit(-1)
    
    if args['microseconds']:
        if int(args['microseconds']) < 0:
            args['microseconds'] = '0'
        
        if int(args['microseconds']) > 2100:
            args['microseconds'] = '2100'
    
    if args['duty']:
        if int(args['duty']) < 0:
            args['duty'] = '0'
        
        if int(args['duty']) > 100:
            args['duty'] = '100'
    
    
    try:
        ch = int(args["channel"])
    except:
        print(f"[ERROR] Invalid Channel Value: {args['channel']}.")
        exit(-3)
    
    if ch < 5 or ch > 16:
        print("[ERROR] Invalid channel value! Must be between 5-16.")
        exit(-2)

    return args

if host == "voxl2":
    import voxl_serial
    dev = voxl_serial.VoxlSerialPort()
    dev.port = '/dev/tty-slpi-7'
    dev.baudrate = 420000
    dev.open()
else:
    # Automatically opens serial
    print("Using /dev/ttyUSB0")
    dev = serial.Serial("/dev/ttyUSB0",420000)

args = parse_args()
ch = int(args['channel'])
crc8 = GENERIC_CRC8(poly=CRSF_CRC_POLY)

# PWM command 
data_out = (uint8_t * (FRAME_SIZE + 2))()
data_out[0] = 0xEC  # ADDR -> RECEIVER
data_out[1] = 0x07  # LEN  -> FRAME_SIZE bytes
data_out[2] = 0x32  # TYPE -> COMMAND
data_out[3] = 0xF4  # CMD  -> MODALAI SET PWM VAL
data_out[4] = ch    # CH   -> 'ch #'

# Microseconds or duty cycle 
if args['microseconds']:
    data_out[5] = 0x73  # TYPE  -> 's'
    data_out[6], data_out[7] = str2us(int(args['microseconds']))  # VAL -> microseconds
elif args['duty']:
    data_out[5] = 0x64  # TYPE  -> 'd'
    data_out[6] = 0x00  
    data_out[7] = int(args['duty'])  # VAL -> duty cycle %

# CRC
data_out[8] = crc8.calc(data_out[2:],FRAME_SIZE - 1, 0)  # CRC   -> Calculated value

bytes_out = bytearray(data_out)
print(f'\nSending {len(data_out)} bytes: {list(data_out)}')
hex_string = ' '.join([format(byte, '02x') for byte in bytes_out])
print(f"{hex_string.upper()}")
dev.write(bytes_out)
dev.close()
