#!/usr/bin/python3
import argparse
import subprocess
import os
import time

BASE_DIR = '/usr/share/modalai/voxl-elrs'
TOOLS_DIR = BASE_DIR + "/tools"

#################
### Utilities ###
#################

def parse_args():
	"""
	Parse and validate args
	"""
	parser = argparse.ArgumentParser(add_help=False)
	parser.add_argument('-f', '--firmware-file', help='Firmware file to flash', required=True, type=str)
	parser.add_argument('-d', '--device', help='Serial port', default='/dev/slpi-uart-7', required=False, type=str)
	return vars(parser.parse_args())

# Check if a service is running
def service_running(service_name):
    """
    Checks if a systemd service is running or not

    @service_name: name of service to check
    @return: True/False depending on service state
    """
    cmd = ['systemctl', 'is-active', service_name]
    result = subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if result.returncode == 0:
        return True
    else:
        return False
    
# Check if a service is enabled
def service_enabled(service_name):
    """
    Checks if a systemd service is enabled or not

    @service_name: name of service to check
    @return: True/False depending on service state
    """
    cmd = ['systemctl', 'is-enabled', service_name]
    result = subprocess.run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if result.returncode == 0:
        return True
    else:
        return False

# Toggle voxl-px4 on or off
def toggle_voxl_px4(enabled=None, running=None):
	"""
	Toggle the state of voxl-px4. Needs to be off when interactive with the receiver.

	@enabled: Enable/Disable state of voxl-px4
	@running: Running state of voxl-px4
	@return: Returns the state of voxl-px4.
	"""
	
	start_voxl_px4_command = ['systemctl', 'start', 'voxl-px4']
	stop_voxl_px4_command = ['systemctl', 'stop', 'voxl-px4']
	enable_voxl_px4_command = ['systemctl', 'enable', 'voxl-px4']
	disable_voxl_px4_command = ['systemctl', 'disable', 'voxl-px4']

	# Case when we want to just disable voxl-px4
	if enabled == None:
		if service_enabled("voxl-px4"):
			# _print(f"[INFO] Disabling voxl-px4.", QUIET)
			enabled = True
			subprocess.run(disable_voxl_px4_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
		else:
			enabled = False
	elif enabled:
		# _print(f"[INFO] Enabling voxl-px4.", QUIET)
		subprocess.run(enable_voxl_px4_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
	elif enabled == False:
		pass
		# _print(f"[INFO] Leaving voxl-px4 disabled.", QUIET)

	# Case when we want to just stop voxl-px4
	if running == None:
		if service_running("voxl-px4"):
			# _print(f"[INFO] Stopping voxl-px4.", QUIET)
			running = True
			subprocess.run(stop_voxl_px4_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
		else:
			running = False
	elif running:
		# _print(f"[INFO] Restarting voxl-px4.", QUIET)
		subprocess.run(start_voxl_px4_command, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
	elif running == False:
		pass
		# _print(f"[INFO] Leaving voxl-px4 disabled.", QUIET)

	return enabled, running

if __name__ == "__main__":
    args = parse_args()

    # Turn off PX4 so we can use slpi uart
    toggle_voxl_px4()
    time.sleep(0.5)
    
    # Put receiver into bootloader mode
    bootloader_command = f"python3 {TOOLS_DIR}/bootloader.py"
    subprocess.run(bootloader_command.split(), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    time.sleep(0.05)
    
    # Flash receiver 
    os.chdir("/usr/share/modalai/voxl-esc-tools/")
    result = subprocess.run(["python3","/usr/share/modalai/voxl-esc-tools/voxl-esc-upload-firmware.py", 
			                "--firmware-file", args['firmware_file'], 
			                "--id", "0",
			                "--skip-firmware-reset", "1",
			                "--device", args['device']],
			                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    exit(result.returncode)